#include <cfloat>
#include <iostream>
#include <fstream>
#include "cylinder.h"
#include <math.h>

#undef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#undef min
#define min(a,b) ((a) < (b) ? (a) : (b))

using namespace std;

enum {SS, SB, SW, ST}; // collision type : chamber-specific

//enum {DOWN = -1, UP = 1};

Cylinder::Cylinder(const char* filename)
    : coef_st_(0.91)
    , coef_sw_(0.2)
    , vr_min_(1e-2)
    , col_st_(0)
    , col_sb_(0)
    , col_sw_(0)
{
    ifstream is(filename);
    if (!is.is_open()) {cerr << "Error reading "<< filename <<endl; exit(1);}
    is >> radius_ >> height_; while(is.get()!='\n');
    radius2_= radius_*radius_; radius_inv_= 1./radius_;

    is >> base_amplitude_ >> base_period_; while(is.get()!='\n');
    base_period_half_= .5*base_period_;
    base_velocity_= 2.*base_amplitude_/base_period_half_;
    base_velocity_inv_= 1./base_velocity_;// if(base_amplitude_)...

    is >> base_z_ >> base_dir_; while(is.get()!='\n');

    if (!is) {cerr << "Error reading "<< filename <<endl; exit(1);}
}

Cylinder::~Cylinder()
{
}

void
Cylinder::advance(double goal)// move cylinder base
{
    static double t0, zt0, zwp;
    if(base_amplitude_) {
        t0= (base_amplitude_ - base_dir_ * base_z_) * base_velocity_inv_;
        zt0= base_z_ + base_dir_ * base_velocity_ *t0;
        while(goal > t0) {
            t0+= base_period_half_;
            base_dir_= -base_dir_;
            zt0= -zt0;
        }
        zwp= zt0 + base_dir_ * base_velocity_ * (goal - t0);
        base_z_= zwp;
    }
}

double
Cylinder::detect(const double *r, const double *v, double best_time, double gravity)
{
    static double t, delta;
    if (height_ && v[Z]>0.) {// detect collision with cylinder top
        t= DBL_MAX;
        double vz2= v[Z]*v[Z];
        delta= vz2 - 2.*(height_ - r[Z])*gravity;
        if (delta>0.) t= (v[Z] - sqrt(delta))/gravity; // division !!!!!!!
        if(t < best_time){
            best_time= t;
            col_type_= ST; //cout << "ST detected "<< t<< endl;
        }
    }

    double b, r2D2, v2D2, dr2D2;
// detect collision with cylinder wall
    b   = v[X]*r[X] + v[Y]*r[Y];
    r2D2= r[X]*r[X] + r[Y]*r[Y];
    v2D2= v[X]*v[X] + v[Y]*v[Y];
    if (r2D2 >(radius2_ + 2e-09)) // particle is outside the cylinder wall
        cout <<"particle at (x,y)=("<<r[X]<<","<<r[Y]<<") "
             <<"with velocity (vx,vy)=("<<v[X]<<","<<v[Y]<<")"<< endl
             <<"has (x^2+y^2="<<r2D2<<")>(R^2="<<radius2_<<") "
             <<"i.e. is outside the cylinder wall"<<endl;

    dr2D2= max(radius2_ - r2D2, 0.);
    t= (-b + sqrt(dr2D2*v2D2 + b*b))/v2D2;
    if (t < best_time) {
        best_time= t;
        col_type_= SW;
    }

// DETECT COLLISION WITH CYLINDER BASE
    static double tpeak, z_tpeak, dz_tpeak, vz_tpeak, dvz_tpeak, tentry;
// detect next peak of sawtooth motion
    tpeak= (base_amplitude_ - base_dir_ * base_z_)*base_velocity_inv_;
    if (base_dir_ < 0) tpeak+= base_period_half_;
    z_tpeak= r[Z] + tpeak * (v[Z] - .5 * gravity * tpeak);
    dz_tpeak= z_tpeak - base_amplitude_;

    double dz0= r[Z] - base_amplitude_;

    tentry = 0.;
    if (dz0 > 0. || dz_tpeak > 0.)// the particle does make an entry/reentry later
        tentry= (v[Z] + sqrt(v[Z]*v[Z] + 2.* dz0 * gravity))/gravity; // DIVISION!

    if (tentry< best_time) {
        while (tentry > tpeak) tpeak += base_period_;
// the collision to detect is now within the cycle preceding "tpeak"
// let's move to the minimum in the middle of this cycle
        tpeak-= base_period_half_;
        z_tpeak= r[Z] + tpeak * (v[Z] - .5 * gravity * tpeak);
        vz_tpeak= v[Z] - gravity * tpeak;
        dz_tpeak= z_tpeak + base_amplitude_;

        if (dz_tpeak < 0 && vz_tpeak < 0) // impact is during descending half-cycle
            dvz_tpeak= vz_tpeak + base_velocity_;
        else                              // impact is during ascending half-cycle
            dvz_tpeak= vz_tpeak - base_velocity_;
        delta= dvz_tpeak*dvz_tpeak + 2.* gravity * dz_tpeak;
        t= tpeak + (dvz_tpeak + sqrt(delta))/gravity; // DIVISION!
        if (t < best_time) {
            best_time= t;
            col_type_= SB;
        }
    }
    return best_time;
}

void
Cylinder::collide(const double *r, double *v)
{
    switch(col_type_) {
        case ST :
            ++col_st_;// if (!(col_st_%10)) echo1(col_st_);
            v[Z]-= (1. + coef_st_) * v[Z];
            break;
        case SB :
            ++col_sb_;// if (!(col_sb_%1000)) echo1(col_sb_);
            v[Z]= 2. * base_velocity_*base_dir_ - v[Z];
            break;
        case SW :
            ++col_sw_;
            double vr= (r[X]*v[X] + r[Y]*v[Y])* radius_inv_;

            if(vr < vr_min_) vr= vr_min_;

            v[X]-= (1. + coef_sw_) * vr * r[X] * radius_inv_;
            v[Y]-= (1. + coef_sw_) * vr * r[Y] * radius_inv_;
            break;
    }
}

int
Cylinder::isInside(const double *r) const
{
    if (height_) if (r[Z] > height_+ 1e-9) return 0;
    if (r[Z] < base_z_- 1e-9) return 0;
    double r2D= sqrt(r[X]*r[X] + r[Y]*r[Y]);
    if (r2D > radius_+ 1e-9) return 0;
    return 1;
}

int
Cylinder::isOutside(const double *r) const
{
    if (height_) if (r[Z] > height_ + 1e-9) return 1;
    if (r[Z] < base_z_- 1e-9) return 1;
    double r2D= sqrt(r[X]*r[X] + r[Y]*r[Y]);
    if (r2D > radius_+ 1e-9) return 1;
    return 0;
}
