#include <iostream>
#include <fstream>
#include <cmath>
#include <cfloat>
#include <cstdlib>
#include "shot.h"

using namespace std;

double
Shot::coef(double velocity)
{
// v_scale_ is the inverse of the threshold velocity
    if (velocity * v_scale_ < 1.) return 1.;// elastic
    return pow(velocity * v_scale_, 0.25);// v^-0.25 decay
}

Shot::Shot(const char* filename)
    : r_(NULL)
    , v_(NULL)
    , tnext_(NULL)
    , partner_(NULL)
    , update_(NULL)
    , chosen_(-1)
    , best_time_(DBL_MAX)

    , collision_(0)
{
    ifstream is(filename);
    if(!is.is_open()){cerr << "Error reading "<< filename << endl; exit(1);}

// constants
    is >> n_ >> diam_ >> mass_; while(is.get()!='\n');
    diam2_= diam_*diam_;
    double v_threshold; is >> v_threshold; while(is.get()!='\n');
    v_scale_= 1./v_threshold;

    if(!is) {cerr << "Error reading " << filename << endl; exit(1);}
    is.close();

    r_= new double*[n_]; v_= new double*[n_];// state vector
    tnext_ = new double[n_];// time to next collision for each particle
    partner_= new int[n_];// its collision partner
    update_ = new int[n_];// collision update flag

    for(int sphere=0; sphere<n_; ++sphere) {
        r_[sphere] = new double[3]; v_[sphere] = new double[3];
        tnext_[sphere] = DBL_MAX;
        partner_[sphere]= -1;
        update_[sphere]= 1;
    }
}

Shot::~Shot()
{
// state vector
    int sphere;
    for (sphere=0; sphere<n_; ++sphere) delete[] r_[sphere]; delete[] r_;
    for (sphere=0; sphere<n_; ++sphere) delete[] v_[sphere]; delete[] v_;

// collision data
    delete[] tnext_; delete[] partner_; delete[] update_;
}

void
Shot::distribute(Chamber& chamber)
{
    int good;
    long seed= 0; srand(seed);
    double r_max= chamber.radius();
    double eta= 0.05; // starting packing fraction
    double z_max= n_* diam2_ * diam_ /(6.*r_max*r_max* eta);

    cout << "random shot distribution started" << endl;
    int i=0;
    while(i<n_) {
        do {
            r_[i][X]= (2.* rand()/double(RAND_MAX) - 1.)*r_max;
            r_[i][Y]= (2.* rand()/double(RAND_MAX) - 1.)*r_max;
            r_[i][Z]= rand()/double(RAND_MAX) * z_max;
        } while(chamber.isOutside(r_[i]));
        good= 1;
        for(int j=0; j<i; ++j) {
            double dx= r_[i][X] - r_[j][X];
            double dy= r_[i][Y] - r_[j][Y];
            double dz= r_[i][Z] - r_[j][Z];
            double dr2= dx*dx + dy*dy + dz*dz;
            if(dr2 < diam2_) good= 0;
        }
        if(good) ++i;
    }
    cout << "random shot distribution finished" << endl;

// pseudo-normal velocity distribution
    double svx = 0., svy = 0., svz = 0.;

    for(i=0; i<n_; ++i) {
        v_[i][X] = 2.*rand()/double(RAND_MAX) - 1.;
        v_[i][Y] = 2.*rand()/double(RAND_MAX) - 1.;
        v_[i][Z] = 2.*rand()/double(RAND_MAX) - 1.;
        svx += v_[i][X]; svy += v_[i][Y]; svz += v_[i][Z];
    }

    svx /= n_; svy /= n_; svz /= n_;
    double vx2 = 0., vy2 = 0., vz2 = 0.;
    for(i=0; i<n_; ++i) {
        v_[i][X] -= svx; vx2 += v_[i][X]*v_[i][X];
        v_[i][Y] -= svy; vy2 += v_[i][Y]*v_[i][Y];
        v_[i][Z] -= svz; vz2 += v_[i][Z]*v_[i][Z];
    }

// we choose kt/m = 1
    double alpha = 1.*sqrt(n_ * 3. / (vx2 + vy2 + vz2));
    for(i=0; i<n_; ++i) {
        v_[i][X]*=alpha; v_[i][Y]*=alpha; v_[i][Z]*=alpha;
    }
}

void
Shot::detect()
{
    static double dx, dy, dz, dvx, dvy, dvz, a, b, c, delta, t;
    for(int i=0; i<n_; ++i) if(update_[i]) {
        for(int j=0; j<n_; ++j) if(j!=i) {
            dx= r_[i][X] - r_[j][X]; dvx= v_[i][X] - v_[j][X];
            dy= r_[i][Y] - r_[j][Y]; dvy= v_[i][Y] - v_[j][Y];
            dz= r_[i][Z] - r_[j][Z]; dvz= v_[i][Z] - v_[j][Z];

            b= dx*dvx + dy*dvy + dz*dvz;
            if(b < 0.) { // the mutual distance is decreasing

// equation to solve in t :
// (dr + t*dv)^2 = sigma^2 : a*t^2 + 2*b*t + c = 0
// with a = dv*dv; b = dr*dv; c = dr*dr - sigma^2

                a= dvx*dvx + dvy*dvy + dvz*dvz;
                c= dx*dx + dy*dy + dz*dz - diam2_;
                if(c < -1e-10)
                    cout <<"negative mutual distance "<< c <<endl
                         <<"between particles "<< i <<" and "<< j <<endl;
                delta= b*b - a*c;
                if(delta > 0.) {
                    t= -(b + sqrt(delta))/a; // we need the SMALLER solution
                    if(t < -1e-4) // this happens only if c is negative
                        cout <<"negative sphere-sphere impact time "<< t <<endl
                             <<"for particles "<< i <<" and "<< j <<endl;
                    if(t < tnext_[i]) { // set next collision for i
                        tnext_[i]= t; partner_[i]= j;
                    }
                    if(t < tnext_[j]) { // set next collision for j
                        tnext_[j]= t; partner_[j]= i;
                    }
                }
            }
        }
    }
    best_time_ = DBL_MAX;
    for (int sphere=0; sphere<n_; ++sphere) if (tnext_[sphere] < best_time_) {
        best_time_= tnext_[sphere];
        chosen_= sphere;
    }
    if (best_time_<0.) cout << "NEGATIVE INTRASHOT IMPACT TIME"<< endl;
}

void
Shot::correct(Chamber& chamber, double gravity)
{
    chamber_impact_= 0;
    for (int sphere=0; sphere<n_; ++sphere){
        double newtime= chamber.detect(r_[sphere], v_[sphere], best_time_, gravity);
        if (newtime < best_time_){
            chosen_= sphere;
            best_time_= newtime;
            chamber_impact_= 1;
        }
    }
}

void
Shot::advance(double time, double gravity)
{
//cout << "time "<< time << endl;
//cout << "chosen "<<chosen_<<" zchosen " << r_[chosen_][Z]<< " vchosen " << v_[chosen_][Z]<< endl;
    for (int i=0; i<n_; ++i){
        r_[i][X]+= time * v_[i][X];
        r_[i][Y]+= time * v_[i][Y];
        r_[i][Z]+= time *(v_[i][Z] - .5* time* gravity);
        v_[i][Z]-= time * gravity;
        tnext_[i]-=time;
    }
    best_time_-=time;
//cout << "zchosen " << r_[chosen_][Z]<< endl;
}

void
Shot::collide(Chamber& chamber)
{
    static double dx, dy, dz, dr2, dvx, dvy, dvz, dv2, a, b;
    int partner = partner_[chosen_];

    if(chamber_impact_) chamber.collide(r_[chosen_], v_[chosen_]);
    else {
        ++collision_;// if (!(collisions_%1)) cout <<collisions_<<endl;
        if(partner < 0 || partner >= n_) {
            cout << "#ERROR sphere-sphere collision number "<<collision_ << endl
                 << "#partner of sphere "<<chosen_<<" has index "<<partner<<" out of range"
                 << endl;
            exit(1);
        }

        dx= r_[chosen_][X] - r_[partner][X];
        dy= r_[chosen_][Y] - r_[partner][Y];
        dz= r_[chosen_][Z] - r_[partner][Z];
        dr2= dx*dx + dy*dy + dz*dz;

        if(fabs(dr2 - diam2_)>1e-6) cout
            << "#WARNING sphere-sphere collision number "<<collision_ <<endl
            << "#spheres "<<chosen_<<" and "<<partner<<" are not in contact"<<endl;

        dvx= v_[chosen_][X] - v_[partner][X];
        dvy= v_[chosen_][Y] - v_[partner][Y];
        dvz= v_[chosen_][Z] - v_[partner][Z];
        dv2= dvx*dvx + dvy*dvy + dvz*dvz;

//        a= 1. + coef(sqrt(dv2));
        a= 1. + .91;

        b= a*.5*(dx*dvx + dy*dvy + dz*dvz)/diam2_;
        dx*= b; dy*= b; dz*= b;
        v_[chosen_][X]-= dx; v_[partner][X]+= dx;
        v_[chosen_][Y]-= dy; v_[partner][Y]+= dy;
        v_[chosen_][Z]-= dz; v_[partner][Z]+= dz;
    }
// common post-impact procedure : set update flags
    for(int sphere=0; sphere< n_; ++sphere){
        update_[sphere]= 0;// this is set to 1 only if...
        if(chosen_==sphere
        || partner==sphere
        || chosen_==partner_[sphere]
        || partner==partner_[sphere]
        ) update_[sphere]= 1;
        if(update_[sphere] == 1){
            tnext_[sphere]= DBL_MAX;
            partner_[sphere]= -1;// OUT OF RANGE
        }
    }
}

void
Shot::test(const Chamber& chamber)
{
    for(int i=0; i<n_; ++i) {
        if(chamber.isOutside(r_[i])) {cout<< "NOT IN CHAMBER"<< endl; exit(1);}
        for(int j=0; j <i; ++j) {
            double dx= r_[i][X] - r_[j][X];
            double dy= r_[i][Y] - r_[j][Y];
            double dz= r_[i][Z] - r_[j][Z];
            double dr2= dx*dx + dy*dy + dz*dz;

            if(dr2 < diam2_ - 1e-9) {cout <<"OVERLAP"<< endl; exit(1);}
        }
    }
}

void
Shot::dumpState(const char* filename)
{
    ofstream os(filename);
    if(!os.is_open()) {cerr <<"Error writing "<<filename<<endl; exit(1);}
    for (int i=0; i< n_; ++i)
        os << r_[i][X] <<" "<< r_[i][Y] <<" "<< r_[i][Z] <<" "
            <<v_[i][X] <<" "<< v_[i][Y] <<" "<< v_[i][Z] << endl;
}
