#ifndef __timer_h__
#define __timer_h__

#include <time.h>

class Timer {
public:
    Timer(const char* keystring = "cpu");

    // Modifier(s)
    // Returns the time elapsed in seconds.
    inline double   elapsed();

    // Returns absolute time in seconds.
    double          absolute();

    // Resets and starts timer. Initial construction also starts timer.
    void            start();

    // Stops the timer. Elapsed time value will not change after the
    // timer is stopped.
    void            stop();

    // Starts timer without resetting.
    void            resume();

    // Operator(s)
    // Another way to return the elapsed time.
    operator double();

private:
    int             real_time_;
    int             isRunning_;
    double          startTime_;
    double          stopTime_;
    double          stopInterval_;
    double          clocks2seconds_;
};

// Inline implementation of ncTimer _____________________________________

inline
Timer::Timer(const char* keystring)
:clocks2seconds_(1./CLOCKS_PER_SEC)
{
    if(!strcmp(keystring,"cpu")) {
        real_time_ = 0;
    }
    else if(!strcmp(keystring,"real")) {
        real_time_ = 1;
    }
    else{
        cerr<<"Timer: keystring \""<< keystring <<"\" is unknown" <<endl;
        exit(1);
    }
    start();
}

#ifdef _WIN32
#include <windows.h>
inline double
Timer::absolute()
{
    // time() returns system time in sec since 01/01/1970
    if(real_time_) {
        time_t t;
        return  double(time(&t));
    }

    static LARGE_INTEGER qpf_freq;
    static int qpf_initialized=0;

    if(!qpf_initialized) {
        if(!QueryPerformanceFrequency(&qpf_freq)){
            cerr<<"Timer: could not obtain frequency" <<endl;
            exit(1);
        }
        qpf_initialized = 1;
    }
    LARGE_INTEGER   y;
    QueryPerformanceCounter(&y);
    double  tm = double(y.LowPart) / double(qpf_freq.LowPart)
                + double(y.HighPart) / double(qpf_freq.LowPart)
                * double(MAXDWORD);
    return tm;
}
#else
inline double
Timer::absolute()
{
    // time() returns system time in sec since 01/01/1970
    if(real_time_) {
        return  double(time(0));
    }
    return double(clock()*clocks2seconds_);
}
#endif

inline double
Timer::elapsed()
{
    if(isRunning_)
        return  absolute() - startTime_ - stopInterval_;
    else
        return  stopTime_ - startTime_ - stopInterval_;
}

inline void
Timer::start()
{
    stopInterval_ = 0.0;
    isRunning_ = 1;
    startTime_ = absolute();
}

inline void
Timer::stop()
{
    if(isRunning_) {
        stopTime_ = absolute();
        isRunning_ = 0;
    }
}

inline void
Timer::resume()
{
    if(isRunning_) return;

    double curTime = absolute();
    stopInterval_ += curTime - stopTime_;
    stopTime_ = curTime;
    isRunning_ = 1;
}

inline
Timer::operator double()
{
    return  elapsed();
}

#endif /* __timer_h__ */
